/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.cfa;

import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.GraphUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Given a network topology, and the offered traffic, this algorithm first routes 
 * the traffic according to the shortest path (in number of traversed links or 
 * in number of traversed km), and then fixes the capacities so that the utilization 
 * in all the links is equal to a user-defined given value.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.2, September 2015
 */
public class CFA_shortestPathFixedUtilization implements IAlgorithm
{
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		final int N = netPlan.getNumberOfNodes();
		final int E = netPlan.getNumberOfLinks();
		final int D = netPlan.getNumberOfDemands();
		if (N == 0 || E == 0 || D == 0)	throw new Net2PlanException("This algorithm requires a topology with links and a demand set");

		/* Initialize some variables */
		final double cg = Double.parseDouble(algorithmParameters.get("cg"));
		final String shortestPathType = algorithmParameters.get("shortestPathType");
		if ((cg <= 0) || (cg > 1)) throw new Net2PlanException("Cg parameter must be positive, and below or equal to 1");
		
		if (!shortestPathType.equalsIgnoreCase("km") && !shortestPathType.equalsIgnoreCase("hops"))
			throw new Net2PlanException("Wrong shortestPathType parameter");

		/* The cost of each link for computing the shortest path depends on the user defined parameter "shortestPathType" */
		Set<Long> linkIds = netPlan.getLinkIds();
		Map<Long, Pair<Long, Long>> linkMap = netPlan.getLinkMap();
		Map<Long, Double> linkCostMap = shortestPathType.equalsIgnoreCase("hops") ? DoubleUtils.ones(linkIds) : netPlan.getLinkLengthInKmMap();

		/* Update the netPlan object with the new routes */
		netPlan.removeAllRoutingInformation();
		netPlan.setRoutingType(Constants.RoutingType.SOURCE_ROUTING);

		Set<Long> demandIds = netPlan.getDemandIds();
		Map<Long, Double> h_d = netPlan.getDemandOfferedTrafficMap();
		for (long demandId : demandIds)
		{
			/* compute the shortest path for this demand */
			long ingressNodeId = netPlan.getDemandIngressNode(demandId);
			long egressNodeId = netPlan.getDemandEgressNode(demandId);
			double trafficValue = h_d.get(demandId);
			List<Long> seqLinks = GraphUtils.getShortestPath(linkMap, ingressNodeId, egressNodeId, linkCostMap);
			if (seqLinks.isEmpty())  continue;

			/* Add the route, no protection segments assigned to the route */
			netPlan.addRoute(demandId, trafficValue, seqLinks, null);
		}

		/* For each link, set the capacity as the one which fixes the utilization to the given value */
		Map<Long, Double> y_e = netPlan.getLinkCarriedTrafficMap();
		for (long linkId : linkIds) netPlan.setLinkCapacity(linkId, y_e.get(linkId) / cg);

		return "Ok!";
	}

	@Override
	public String getDescription()
	{
		return "Routes the traffic according to the shortest path, and then fixes the link capacities to match a given congestion figure";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("cg", "0.6", "Fixed link utilization"));
		algorithmParameters.add(Triple.of("shortestPathType", "#select# hops km", "Criteria to compute the shortest path. Valid values: 'hops' or 'km'"));
		
		return algorithmParameters;
	}
}